using System.Reflection;
using System.Security.Cryptography;
using System.Text;
using System.Text.RegularExpressions;

namespace Cogwheel;

/// <summary>
/// Manages commands for the Cogwheel system.
/// </summary>
public class CommandsManager
{
    /// <summary>
    /// Gets or sets the console used by the CommandsManager.
    /// </summary>
    protected virtual ICogwheelConsole CogwheelConsole { get; set; }

    /// <summary>
    /// Gets or sets the pattern used to parse commands.
    /// </summary>
    protected virtual string CommandPattern { get; set; } = "(?<val>(\\([^\\)]+\\)))|\"(?<val>[^\"]+)\"|'(?<val>[^']+)'|(?<val>[^\\s]+)";

    /// <summary>
    /// Gets or sets the list of assemblies to search for commands.
    /// </summary>
    protected virtual List<Assembly> Assemblies { get; set; } = new();

    /// <summary>
    /// Gets or sets the dictionary of commands.
    /// </summary>
    public virtual Dictionary<string, ICommand> Commands { get; set; } = new();

    /// <summary>
    /// Gets or sets the dictionary of custom parsers for command parameters.
    /// </summary>
    public virtual Dictionary<Type, Func<string, object>> CustomParsers { get; set; } = new();

    /// <summary>
    /// Gets or sets the current context object.
    /// </summary>
    public virtual object? CurrentContext { get; set; }

    /// <summary>
    /// Gets or sets the GUID of the current context.
    /// </summary>
    public virtual Guid? CurrentContextGuid { get; set; }

    /// <summary>
    /// Gets or sets the dictionary of registered object instances.
    /// </summary>
    public virtual Dictionary<Guid, object> RegisteredObjectInstances { get; set; } = new();

    /// <summary>
    /// Gets or sets the dictionary of registered object GUIDs.
    /// </summary>
    public virtual Dictionary<object, Guid> RegisteredObjectGuids { get; set; } = new();

    /// <summary>
    /// Initializes the CommandsManager with the specified console.
    /// </summary>
    /// <param name="console">The console to use.</param>
    public virtual void Initialize(ICogwheelConsole console)
    {
        CogwheelConsole = console;
        RefreshCommandsList();
        CogwheelConsole.Log($"CommandsManager initialized, Commands found: {Commands.Count}");
    }

    /// <summary>
    /// Gets the command and arguments from the input string.
    /// </summary>
    /// <param name="input">The input string.</param>
    /// <returns>A tuple containing the command and arguments, or null if the command is not found.</returns>
    public virtual (ICommand, List<object>)? GetCommandAndArgsFromString(string input)
    {
        var splitString = Regex.Matches(input, CommandPattern).Select(m => m.Groups["val"].Value).ToArray();
        var commandName = splitString[0];
        List<object> args = new();
        for (var splitIndex = 1; splitIndex < splitString.Length; splitIndex++)
        {
            args.Add(splitString[splitIndex]);
        }

        if (Commands.TryGetValue(commandName, out var command))
        {
            return (command, args);
        }

        return null;
    }

    /// <summary>
    /// Runs the command specified by the input string.
    /// </summary>
    /// <param name="input">The input string.</param>
    /// <param name="context">The context object, if any.</param>
    /// <returns>True if the command was executed successfully, otherwise false.</returns>
    public virtual bool RunCommand(string input, object? context = null)
    {
        if (string.IsNullOrWhiteSpace(input)) return false;
        var command = GetCommandAndArgsFromString(input);
        if (command is null)
        {
            CogwheelConsole.LogError($"Command not found: {input}");
            return false;
        }

        if (command.Value.Item1.Method.IsStatic)
        {
            return ExecuteCommand(null, command.Value);
        }

        if (context is not null && IsCommandContextValid(command.Value.Item1, context))
        {
            return ExecuteCommand(context, command.Value);
        }

        if (CurrentContext is not null && IsCommandContextValid(command.Value.Item1, CurrentContext))
        {
            return ExecuteCommand(CurrentContext, command.Value);
        }

        if (command.Value.Item1.Method.DeclaringType != null)
        {
            context = GetFirstValidRegisteredContext(command.Value.Item1.Method.DeclaringType);
            if (context is null)
            {
                CogwheelConsole.LogError($"No context of type '{command.Value.Item1.Method.DeclaringType}' found");
                return false;
            }
        }
        else
        {
            CogwheelConsole.LogError($"Command '{command.Value.Item1.Name}' has no declaring type");
            return false;
        }

        if (IsCommandContextValid(command.Value.Item1, context))
        {
            return ExecuteCommand(context, command.Value);
        }

        return false;
    }

    /// <summary>
    /// Gets the usage information for the specified command.
    /// </summary>
    /// <param name="command">The command.</param>
    /// <returns>The usage information for the command.</returns>
    public virtual string GetCommandUsage(ICommand command)
    {
        string output = "";
        if (!command.Method.IsStatic)
        {
            output += $"|{command.Method.DeclaringType?.FullName}| ";
        }
        else
        {
            output += "|Global| ";
        }

        output += $"{command.Name}: ";

        if (!string.IsNullOrWhiteSpace(command.Description))
        {
            output += $"{command.Description} ";
        }

        string paramUsage = string.Join(" ",
            command.Method.GetParameters().Select(param =>
                $"<{(param.IsDefined(typeof(ParamArrayAttribute)) ? "params " : "")}{param}>"));

        if (!string.IsNullOrWhiteSpace(paramUsage))
        {
            output += $"- {paramUsage}";
        }

        return output;
    }

    /// <summary>
    /// Executes the specified command with the given arguments.
    /// </summary>
    /// <param name="obj">The object on which to execute the command.</param>
    /// <param name="command">The command and arguments.</param>
    /// <returns>True if the command was executed successfully, otherwise false.</returns>
    public virtual bool ExecuteCommand(object? obj, (ICommand command, List<object> args) command)
    {
        var parameters = command.command.Method.GetParameters();
        for (var parameterIndex = 0; parameterIndex < parameters.Length; parameterIndex++)
        {
            if (parameters[parameterIndex].IsDefined(typeof(ParamArrayAttribute), false))
            {
                var paramList = Activator.CreateInstance(typeof(List<>).MakeGenericType(parameters[parameterIndex].ParameterType.GetElementType()!));
                for (var argIndex = parameterIndex; argIndex < parameters.Length; argIndex++)
                {
                    if (TryParseParameter(parameters[parameterIndex].ParameterType.GetElementType(),
                            (string)command.args[argIndex], out var val))
                    {
                        paramList.GetType().GetMethod("Add")?.Invoke(paramList, [val]);
                    }
                    else
                    {
                        CogwheelConsole.LogError($"Format exception: could not parse '{command.args[parameterIndex]}' as '{parameters[parameterIndex].ParameterType}'");
                        return false;
                    }
                }

                command.args = command.args.Take(parameterIndex).ToList();
                command.args.Add(paramList.GetType().GetMethod("ToArray")?.Invoke(paramList, null));
            }
            else if (parameterIndex >= command.args.Count)
            {
                if (parameters[parameterIndex].IsOptional)
                {
                    command.args.Add(Type.Missing);
                }
                else
                {
                    CogwheelConsole.LogError($"Not enough args passed. Usage: {GetCommandUsage(command.command)}");
                    return false;
                }
            }
            else
            {
                if (TryParseParameter(parameters[parameterIndex].ParameterType, (string)command.args[parameterIndex], out var val))
                {
                    command.args[parameterIndex] = val;
                }
                else
                {
                    CogwheelConsole.LogError($"Format exception: could not parse '{command.args[parameterIndex]}' as '{parameters[parameterIndex].ParameterType}'");
                    return false;
                }
            }
        }

        try
        {
            command.command.Method.Invoke(obj, command.args.ToArray());
            return true;
        }
        catch (Exception e)
        {
            CogwheelConsole.LogError(e.Message);
            throw;
        }
    }

    /// <summary>
    /// Tries to parse the parameter string into the specified parameter type.
    /// </summary>
    /// <param name="parameterType">The type of the parameter.</param>
    /// <param name="parameterString">The parameter string.</param>
    /// <param name="parsedValue">The parsed value.</param>
    /// <returns>True if the parameter was parsed successfully, otherwise false.</returns>
    public virtual bool TryParseParameter(Type parameterType, string parameterString, out object? parsedValue)
    {
        if (parameterType == typeof(string))
        {
            parsedValue = parameterString;
            return true;
        }

        if (CustomParsers.ContainsKey(parameterType))
        {
            try
            {
                parsedValue = CustomParsers[parameterType].Invoke(parameterString);
                return true;
            }
            catch (Exception e)
            {
                CogwheelConsole.LogError(e.Message);
                parsedValue = null;
                return false;
            }
        }
        else
        {
            var parseMethod = parameterType.GetMethod("Parse", new[] { typeof(string) });

            if (parseMethod is not null)
            {
                try
                {
                    parsedValue = parseMethod.Invoke(null, new object[] { parameterString });
                    return true;
                }
                catch (Exception e)
                {
                    CogwheelConsole.LogError(e.Message);
                    parsedValue = null;
                    return false;
                }
            }
        }

        parsedValue = null;
        return false;
    }

    /// <summary>
    /// Gets the command by its name.
    /// </summary>
    /// <param name="commandName">The name of the command.</param>
    /// <returns>The command, or null if not found.</returns>
    public virtual ICommand? GetCommandByName(string commandName)
    {
        return Commands.GetValueOrDefault(commandName);
    }

    /// <summary>
    /// Registers the specified object with the CommandsManager.
    /// </summary>
    /// <param name="obj">The object to register.</param>
    public virtual void RegisterObject(object obj)
    {
        string seed = obj.GetType().FullName + GetObjectPropertiesHash(obj);
        using (MD5 md5 = MD5.Create())
        {
            byte[] hash = md5.ComputeHash(Encoding.UTF8.GetBytes(seed));
            Guid uniqueId = new Guid(hash.Take(16).ToArray());

            if (!RegisteredObjectInstances.ContainsKey(uniqueId))
            {
                RegisteredObjectInstances[uniqueId] = obj;
                RegisteredObjectGuids[obj] = uniqueId;
                CogwheelConsole.Log($"Registered object with ID {uniqueId}");
            }
            else
            {
                CogwheelConsole.LogWarning($"Object with ID {uniqueId} is already registered");
            }
        }
    }

    /// <summary>
    /// Checks if the command context is valid.
    /// </summary>
    /// <param name="command">The command.</param>
    /// <param name="context">The context object, if any.</param>
    /// <returns>True if the context is valid, otherwise false.</returns>
    public virtual bool IsCommandContextValid(ICommand command, object? context = null)
    {
        if (context is not null)
        {
            return command.Method.DeclaringType == context.GetType();
        }
        return command.Method.DeclaringType == CurrentContext?.GetType();
    }

    /// <summary>
    /// Sets the current context to the specified object.
    /// </summary>
    /// <param name="context">The context object.</param>
    /// <returns>True if the context was set successfully, otherwise false.</returns>
    public virtual bool SetContext(object context)
    {
        CurrentContext = context;
        if (!RegisteredObjectGuids.ContainsKey(context))
        {
            RegisterObject(context);
            return true;
        }
        CurrentContextGuid = RegisteredObjectGuids[context];
        return true;
    }

    /// <summary>
    /// Sets the current context to the object with the specified GUID.
    /// </summary>
    /// <param name="guid">The GUID of the context object.</param>
    /// <returns>True if the context was set successfully, otherwise false.</returns>
    public virtual bool SetContext(Guid guid)
    {
        if (RegisteredObjectInstances.ContainsKey(guid))
        {
            CurrentContext = RegisteredObjectInstances[guid];
            CurrentContextGuid = guid;
        }
        else
        {
            CogwheelConsole.LogError($"No object registered with ID {guid}");
            return false;
        }

        return true;
    }

    /// <summary>
    /// Sets the current context to the object with the specified GUID string.
    /// </summary>
    /// <param name="guidString">The GUID string of the context object.</param>
    /// <returns>True if the context was set successfully, otherwise false.</returns>
    public virtual bool SetContext(string guidString)
    {
        if (Guid.TryParse(guidString, out var guid))
        {
            return SetContext(guid);
        }

        CogwheelConsole.LogError($"Could not parse '{guidString}' as a GUID");
        return false;
    }

    /// <summary>
    /// Gets the first valid registered context of the specified type.
    /// </summary>
    /// <param name="type">The type of the context.</param>
    /// <returns>The first valid registered context, or null if not found.</returns>
    public virtual object? GetFirstValidRegisteredContext(Type type)
    {
        return RegisteredObjectInstances.Values.FirstOrDefault(obj => obj.GetType() == type);
    }

    /// <summary>
    /// Gets the GUID of the specified context object.
    /// </summary>
    /// <param name="context">The context object, if any.</param>
    /// <returns>The GUID of the context object.</returns>
    public virtual Guid GetGuidFromContext(object? context = null)
    {
        if (context is not null)
        {
            return RegisteredObjectGuids[context];
        }

        return CurrentContextGuid ?? Guid.Empty;
    }
    
    /// <summary>
    /// Adds the specified assembly to the list of assemblies used to search for commands.
    /// </summary>
    /// <param name="assembly"></param>
    public void AddAssembly(Assembly assembly)
    {
        if (!Assemblies.Contains(assembly))
        {
            Assemblies.Add(assembly);
            RefreshCommandsList();
        }
        else
        {
            CogwheelConsole.LogWarning($"Tried to add {assembly.FullName} but it is already registered.");
        }
    }

    /// <summary>
    /// Removes the specified assembly from the list of assemblies used to search for commands.
    /// </summary>
    /// <param name="assembly"></param>
    public void RemoveAssembly(Assembly assembly)
    {
        if (Assemblies.Contains(assembly))
        {
            Assemblies.Remove(assembly);
            RefreshCommandsList();
        }
        else
        {
            CogwheelConsole.LogWarning($"Tried to remove {assembly.FullName} but it is not registered.");
        }
    }
    
    /// <summary>
    /// Clears the list of assemblies used to search for commands.
    /// </summary>
    public void ClearAssemblies()
    {
        Assemblies.Clear();
        RefreshCommandsList();
    }
    
    /// <summary>
    /// Adds a custom parser for the specified type.
    /// </summary>
    /// <param name="type"></param>
    /// <param name="parser"></param>
    public void AddCustomParser(Type type, Func<string, object> parser)
    {
        if (!CustomParsers.ContainsKey(type))
        {
            CustomParsers[type] = parser;
        }
        else
        {
            CogwheelConsole.LogWarning($"Tried to add a custom parser for {type.FullName} but it is already registered.");
        }
    }

    /// <summary>
    /// Gets the hash of the properties of the specified object.
    /// </summary>
    /// <param name="obj">The object.</param>
    /// <returns>The hash of the object's properties.</returns>
    protected virtual string GetObjectPropertiesHash(object obj)
    {
        var properties = obj.GetType().GetProperties();
        var sb = new StringBuilder();

        foreach (var prop in properties)
        {
            var value = prop.GetValue(obj)?.ToString() ?? string.Empty;
            sb.Append(value);
        }

        return sb.ToString();
    }

    /// <summary>
    /// Refreshes the list of commands by scanning the assemblies.
    /// </summary>
    protected virtual void RefreshCommandsList()
    {
        foreach (var assembly in Assemblies)
        {
            foreach (var type in assembly.GetTypes())
            {
                var methods = type.GetMethods(BindingFlags.NonPublic | BindingFlags.Static | BindingFlags.Public | BindingFlags.Instance);

                foreach (var method in methods)
                {
                    var attributes = method.GetCustomAttributes<CommandAttribute>();

                    foreach (var attribute in attributes)
                    {
                        var commandName = attribute.Name ?? method.Name;
                        var newCommand = new Command(commandName, attribute.Description, method);
                        Commands.TryAdd(newCommand.Name, newCommand);
                    }
                }
            }
        }
    }
}