namespace Cogwheel;

using System;

/// <summary>
/// Default implementation of the Cogwheel console.
/// </summary>
public class DeafultCogwheelConsole : ICogwheelConsole
{
    /// <summary>
    /// Gets or sets the opening message of the console.
    /// </summary>
    public string OpeningMessage { get; set; } = "** COGWHEEL CONSOLE **";

    /// <summary>
    /// Gets or sets a value indicating whether the console is running.
    /// </summary>
    public bool IsRunning { get; set; }

    /// <summary>
    /// Gets or sets the commands manager associated with the console.
    /// </summary>
    public CommandsManager CommandsManager { get; set; }

    /// <summary>
    /// Initializes the console with the specified commands manager.
    /// </summary>
    /// <param name="commandsManager">The commands manager to use.</param>
    public void Initialize(CommandsManager commandsManager)
    {
        CommandsManager = commandsManager;
        CommandsManager.RegisterObject(this);

        Write(OpeningMessage);

        IsRunning = true;
        while (IsRunning)
        {
            if (CommandsManager.CurrentContext != null)
            {
                Console.Write($"({CommandsManager.CurrentContext.GetType().FullName})> ");
            }
            else
            {
                Console.Write("(Global)> ");
            }
            string input = Console.ReadLine();
            Console.WriteLine();
            CommandsManager.RunCommand(input);
            Console.WriteLine();
        }
    }

    /// <summary>
    /// Logs a message to the console.
    /// </summary>
    /// <param name="message">The message to log.</param>
    public void Log(string message)
    {
        Console.WriteLine($"[COGWHEEL] {message}");
    }

    /// <summary>
    /// Logs an error message to the console.
    /// </summary>
    /// <param name="message">The error message to log.</param>
    public void LogError(string message)
    {
        Console.WriteLine($"[COGWHEEL ERROR] {message}");
    }

    /// <summary>
    /// Logs a warning message to the console.
    /// </summary>
    /// <param name="message">The warning message to log.</param>
    public void LogWarning(string message)
    {
        Console.WriteLine($"[COGWHEEL WARNING] {message}");
    }

    /// <summary>
    /// Writes a message to the console.
    /// </summary>
    /// <param name="message">The message to write.</param>
    public void Write(string message)
    {
        Console.WriteLine(message);
    }

    /// <summary>
    /// Clears the console.
    /// </summary>
    public void ClearConsole()
    {
        Console.Clear();
    }

    /// <summary>
    /// Exits the console.
    /// </summary>
    public void Exit()
    {
        IsRunning = false;
    }
}